/*
Drawboard - Java applet used to make graphical teleconferences
Copyright (C) 2001  Tomek "TomasH" Zielinski, tomash@fidonet.org.pl

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package drawboard;

import java.util.*;
import java.util.zip.*;
import java.lang.Math;
import java.io.Serializable;

// =============================================================================
// ===========================================================================
// Current image, maintained and distributed to new clients by server class
public class Archive implements Shape, Serializable{

  byte parcel[];   // zipped bitmap
  short DIM_X;      // server image x dim
  short DIM_Y;      // server image y dim
  public int ID=0;  // sender's ID  (0 = server message)

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // Create current image package
  Archive (int[] what, short x, short y){
    DIM_X=x;
    DIM_Y=y;
    int offset=what.length;
    parcel=new byte[offset*3];
    for (int i=0; i<what.length; i++){
      parcel[i]=(byte)((what[i]>>16)&0xff);           // R
      parcel[i+offset]=(byte)((what[i]>>8)&0xff);     // G
      parcel[i+offset+offset]=(byte)(what[i]&0xff);   // B
    }


    byte uwaga[]=new byte[100];
    for (int t=0; t<100; t++)
      uwaga[t]=(byte)(t%255);


    // pack data
    Deflater def=new Deflater();
    def.setInput(parcel,0,parcel.length);  // Deflater.setInput(byte[]) has bug!
    def.finish();
    int bytesDone = def.deflate(parcel);

    // and prepare to transfer
    byte temp[]=new byte[bytesDone];
    for (int i=0; i<bytesDone; i++)temp[i]=parcel[i];
    parcel=new byte[bytesDone];
    for (int i=0; i<bytesDone; i++)parcel[i]=temp[i];
    temp=null;
    def.end();
    System.gc();
  }

  // Draw current image on client's board, when it receives parcel
  public void draw(int[] matrix, short width){
    byte temp[]=new byte[matrix.length*3];

    // unpack data...
    Inflater inf=new Inflater();
    inf.setInput(parcel);
    inf.finished();
    try{
      inf.inflate(temp);
    }catch(DataFormatException e){
      System.out.println("Wrong archive data received "+e);
      return;
    };
    inf.end();

    // ... and copy to the buffre
    int offset=matrix.length;
    for (int i=0; i<offset; i++)
      matrix[i]=
        temp[i+offset+offset]&0xff |
        (((temp[i+offset]&0xff)<<8)) |
        (((temp[i]&0xff)<<16)) |
        (((0xff)<<24));
  };

  // Not used in this class
  public void setx2y2(short x2, short y2){};
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){}
}


// =============================================================================
// ===========================================================================
// A single "drawing" interface
interface Shape {
  public void setID(int i);
  public int getID();
  public void draw(int[] matrix, short width);  // draw themselves
  public void setx2y2(short x2, short y2);      // modify current object
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2);
}

// =============================================================================
// ===========================================================================
// Pinger pseudoshape, contains only people counter information
class Pinger implements Shape, Serializable{
  short people;
  Pinger(short i){people=i;}
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){}
  public void setx2y2(short xx2, short yy2){}
  public void draw(int[] matrix, short width){}
  public int ID=0;

  // Get unique ID
  public int getID(){return ID;}

  // Set unique ID
  public void setID(int i){ID=i;}
}

// =============================================================================
// ===========================================================================
// Reset screen pseudoshape
class Reset implements Shape, Serializable{
  int RGB;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  public Reset(int cc){
    RGB=cc;
  }

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){}

  // =============================================================================
  public void setx2y2(short xx2, short yy2){}

  // =============================================================================
  public void draw(int[] matrix, short width){
    for (int i=0; i<matrix.length; i++)
      matrix[i]=RGB;
  }
}


// ===========================================================================
// =============================================================================
// Single line shape
class Line implements Shape, Serializable{
  short x1, y1, x2, y2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  Line(){};

  // =============================================================================
  Line(short X1, short Y1, int c, short t){
    x1=X1;
    y1=Y1;
    RGB=c;
    thick=t;
  }

  // =============================================================================
  // Light one pixel
  static void pixel(int[] matrix, short width, int x, int y, int color){
    if (x<width && x>=0 && y>=0 && y<matrix.length/width)
      matrix[y*width+x]=color;
  }

  // =============================================================================
  // Light screen "dot", diameter-thick circle
  static final void dot(int[] matrix, short width, short x, short y, short thick, int color){
    int height=matrix.length/width;
    switch (thick){

      case 1:
        pixel(matrix,width,x,y,color);              //   *  - shape
        break;

      case 2:
        pixel(matrix,width,x,y,color);              //   X*  - shape
        pixel(matrix,width,x+1,y,color);
        pixel(matrix,width,x,y+1,color);
        pixel(matrix,width,x+1,y+1,color);
        break;

      case 3:
        pixel(matrix,width,x,y,color);
        pixel(matrix,width,x+1,y,color);      //    *
        pixel(matrix,width,x,y+1,color);      //   *X*
        pixel(matrix,width,x-1,y,color);      //    *
        pixel(matrix,width,x,y-1,color);
        break;

      case 4:
        pixel(matrix,width,x,y,color);
        pixel(matrix,width,x,y-1,color);    //  **
        pixel(matrix,width,x+1,y-1,color);  // *X**
        pixel(matrix,width,x-1,y,color);    // ****
        pixel(matrix,width,x+1,y,color);    //  **
        pixel(matrix,width,x+2,y,color);
        pixel(matrix,width,x-1,y+1,color);
        pixel(matrix,width,x,y+1,color);
        pixel(matrix,width,x+1,y+1,color);
        pixel(matrix,width,x+2,y+1,color);
        pixel(matrix,width,x,y+2,color);
        pixel(matrix,width,x+1,y+2,color);
        break;

      case 5:
        pixel(matrix,width,x,y,color);
        pixel(matrix,width,x,y-2,color);
        pixel(matrix,width,x-1,y-1,color);    //    *
        pixel(matrix,width,x,y-1,color);      //   ***
        pixel(matrix,width,x+1,y-1,color);    //  **X**
        pixel(matrix,width,x-2,y,color);      //   ***
        pixel(matrix,width,x-1,y,color);      //    *
        pixel(matrix,width,x,y,color);
        pixel(matrix,width,x+1,y,color);
        pixel(matrix,width,x+2,y,color);
        pixel(matrix,width,x-1,y+1,color);
        pixel(matrix,width,x,y+1,color);
        pixel(matrix,width,x+1,y+1,color);
        pixel(matrix,width,x,y+2,color);
        break;

      case 6:
        pixel(matrix,width,x,y,color);
        pixel(matrix,width,x,y-2,color);
        pixel(matrix,width,x+1,y-2,color);
        pixel(matrix,width,x-1,y-1,color);    //    **
        pixel(matrix,width,x,y-1,color);      //   ****
        pixel(matrix,width,x+1,y-1,color);    //  **X***
        pixel(matrix,width,x+2,y-1,color);    //  ******
        pixel(matrix,width,x-2,y,color);      //   ****
        pixel(matrix,width,x-1,y,color);      //    **
        pixel(matrix,width,x+1,y,color);
        pixel(matrix,width,x+2,y,color);
        pixel(matrix,width,x+3,y,color);
        pixel(matrix,width,x-2,y+1,color);
        pixel(matrix,width,x-1,y+1,color);
        pixel(matrix,width,x,y+1,color);
        pixel(matrix,width,x+1,y+1,color);
        pixel(matrix,width,x+2,y+1,color);
        pixel(matrix,width,x+3,y+1,color);
        pixel(matrix,width,x-1,y+2,color);
        pixel(matrix,width,x,y+2,color);
        pixel(matrix,width,x+1,y+2,color);
        pixel(matrix,width,x+2,y+2,color);
        pixel(matrix,width,x,y+3,color);
        pixel(matrix,width,x+1,y+3,color);
    }
  }

  // =============================================================================
  // Used also by another shapes
  public final static void drawthickline(int[] matrix, short width,int RGB,
        short x0, short y0, short x1, short y1, short thick){

    int color=RGB;

    short x = x0, y = y0;
    int dx = x1-x0, dy = y1-y0;   // direction of line

    int sx = (dx > 0 ? 1 : (dx < 0 ? -1 : 0));    // increment or decrement depending on direction of line
    int sy = (dy > 0 ? 1 : (dy < 0 ? -1 : 0));

    if ( dx < 0 ) dx = -dx;    // decision parameters for voxel selection
    if ( dy < 0 ) dy = -dy;
    int ax = 2*dx, ay = 2*dy;
    int decx, decy;

    int max = dx, var = 0;    // determine largest direction component, single-step related variable
    if ( dy > max ) { var = 1; }

    switch ( var ){    // traverse Bresenham line
    case 0:  // single-step in x-direction
      for (decy=ay-dx; /**/; x += sx, decy += ay){
        dot(matrix,width,x,y,thick, color);
          if ( x == x1 ) break;            // take Bresenham step
          if ( decy >= 0 ) { decy -= ax; y += sy; }
        }
        break;
    case 1:  // single-step in y-direction
        for (decx=ax-dy; /**/; y += sy, decx += ax){
        dot(matrix,width,x,y,thick, color);
            if ( y == y1 ) break;            // take Bresenham step
            if ( decx >= 0 ) { decx -= ay; x += sx; }
        }
        break;
    }
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    drawthickline( matrix,width,RGB,x1,y1,x2, y2, thick);
  }

  // =============================================================================
  public void setx2y2(short xx2, short yy2){
    x2=xx2;y2=yy2;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    x1=xx1;
    y1=yy1;
    RGB=cc;
    thick=tt;
    x2=xx2;y2=yy2;
  };

}

// =============================================================================
// ===========================================================================
// Circle (ellipse) shape
class Circle implements Shape, Serializable{
  short x1, y1, x2, y2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  Circle(){};

  // =============================================================================
  Circle(short X, short Y, int c, short t){
    x1=X;
    y1=Y;
    RGB=c;
    thick=t;
  }

  // =============================================================================
  public static void elipsePoints(int[] matrix, final short width, final short xc, final short yc, final short x, final short y, final int RGB, final short thick){
    Line.dot(matrix,width,(short)(xc+x),(short)(yc+y),thick,RGB);
    Line.dot(matrix,width,(short)(xc+x),(short)(yc-y),thick,RGB);
    Line.dot(matrix,width,(short)(xc-x),(short)(yc+y),thick,RGB);
    Line.dot(matrix,width,(short)(xc-x),(short)(yc-y),thick,RGB);
   }


  public static void ellipseDraw(int[] matrix, final short width, final short xc, final short yc, final short a, final short b, final int RGB, final short thick){
    int a2 = a*a;
    int b2 = b*b;

    int x, y, dec;
    for (x = 0, y = b, dec = 2*b2+a2*(1-2*b); b2*x <= a2*y; x++){
      elipsePoints( matrix, width,xc,yc,(short)x,(short)y,RGB,thick);
      if ( dec >= 0 )
          dec += 4*a2*(1-(y--));
      dec += b2*(4*x+6);
    }

    for (x = a, y = 0, dec = 2*a2+b2*(1-2*a); a2*y <= b2*x; y++){
      elipsePoints( matrix, width,xc,yc,(short)x,(short)y,RGB,thick);
      if ( dec >= 0 )
          dec += 4*b2*(1-(x--));
      dec += a2*(4*y+6);
    }

  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    short xx=(short)Math.abs(x1-x2);
    short yy=(short)Math.abs(y1-y2);
    if (xx==00 || yy==00) return;
    short a=0, b=0;
    if (xx<yy){
      a=(short)Math.min(xx,yy);
      b=(short)Math.max(xx,yy);
    }else{
      b=(short)Math.min(xx,yy);
      a=(short)Math.max(xx,yy);
    }
    ellipseDraw(matrix,width,x1,y1,a,b,RGB, thick);
  }


  // =============================================================================
  public void setx2y2(short xx2, short yy2){
    x2=xx2;y2=yy2;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    x1=xx1;
    y1=yy1;
    RGB=cc;
    thick=tt;
    x2=xx2;y2=yy2;
  };

}


// =============================================================================
// ===========================================================================
// Filled circle (ellipse) shape
class FillCircle extends Circle implements Shape, Serializable{
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  FillCircle(){};

  // =============================================================================
  FillCircle(short X, short Y, int c, short t){
    super(X, Y, c, t);
  }

  public static void fillElipsePoints(int[] matrix, int xc, int yc, int x, int y, int RGB, int width){
    int ysize=matrix.length/width;
    for (int i=Math.max(0,xc-x); i<Math.min(width,xc+x); i++){
      if (yc-y>=0)  matrix[(yc-y)*width+i]=RGB;
    }
    for (int i=Math.max(0,xc-x); i<Math.min(width,xc+x); i++){
      if (yc+y<ysize) matrix[(yc+y)*width+i]=RGB;
    }
  }

  public static void drawFillEllipse(int[] matrix, int x1, int y1, int x2, int y2, int RGB, int width){
    int xx=Math.abs(x1-x2);
    int yy=Math.abs(y1-y2);
    if (xx==00 || yy==00) return;
    int a=0, b=0;
    if (xx<yy){
      a=Math.min(xx,yy);
      b=Math.max(xx,yy);
    }else{
      b=Math.min(xx,yy);
      a=Math.max(xx,yy);
    }

    int a2 = a*a;
    int b2 = b*b;

    int x, y, dec;
    for (x = 0, y = b, dec = 2*b2+a2*(1-2*b); b2*x <= a2*y; x++){
      fillElipsePoints( matrix,x1,y1,x,y,RGB,width);
      if ( dec >= 0 )
          dec += 4*a2*(1-(y--));
      dec += b2*(4*x+6);
    }
    for (x = a, y = 0, dec = 2*a2+b2*(1-2*a); a2*y <= b2*x; y++){
      fillElipsePoints( matrix,x1,y1,x,y,RGB,width);
        if ( dec >= 0 )
            dec += 4*b2*(1-(x--));
        dec += a2*(4*y+6);
    }
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    drawFillEllipse( matrix,x1,y1,x2,y2,RGB,width);
  }
}


// =============================================================================
// ===========================================================================
// Box (rectangle) shape
class Box implements Shape, Serializable{
  short x1, y1, x2, y2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  Box(){};

  // =============================================================================
  Box(short X, short Y, int c, short t){
    x1=X;
    y1=Y;
    RGB=c;
    thick=t;
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    Line.drawthickline(matrix,width,RGB,x1,y1,x2,y1, thick);
    Line.drawthickline(matrix,width,RGB,x2,y1,x2,y2, thick);
    Line.drawthickline(matrix,width,RGB,x2,y2,x1,y2, thick);
    Line.drawthickline(matrix,width,RGB,x1,y2,x1,y1, thick);
  }

  // =============================================================================
  public void setx2y2(short xx2, short yy2){
    x2=xx2;y2=yy2;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    x1=xx1;
    y1=yy1;
    RGB=cc;
    thick=tt;
    x2=xx2;y2=yy2;
  };
}


// =============================================================================
// ===========================================================================
// Filled box (rectangle) shape
class FillBox extends Box implements Shape, Serializable{
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  FillBox(){};

  // =============================================================================
  FillBox(short X, short Y, int c, short t){
    super(X, Y, c, t);
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    for (int y=Math.min(super.y1, super.y2); y<Math.max(super.y1, super.y2);y++){
      int offset=y*width;
      for (int x=Math.min(super.x1,super.x2); x<Math.max(super.x1,super.x2);x++){
        matrix[offset+x]=RGB;
      }
    }
  }
}


// ===========================================================================
// =============================================================================
// Freehand shape
class Freehand implements Shape, Serializable{
  final short SIZE=255;
  short x[]=new short[SIZE];
  short y[]=new short[SIZE];
  short counter=2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  Freehand(){};

  // =============================================================================
  Freehand(short X, short Y, int c, short t){
    x[0]=X;
    y[0]=Y;

    RGB=c;
    thick=t;
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    for (short a=0; a<counter-1; a++){
      Line.drawthickline(matrix,width,RGB,x[a],y[a],x[a+1],y[a+1], thick);
    }
  }

  // =============================================================================
  public void setx2y2(short xx2, short yy2){
  };

  // =============================================================================
  boolean nextxy(short xx, short yy){
    x[counter]=xx;
    y[counter]=yy;
    counter++;
    if (counter==SIZE) return true; else return false;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    counter=2;
    x[0]=xx1;
    y[0]=yy1;

    RGB=cc;
    thick=tt;
    x[1]=xx2;
    y[1]=yy2;
  };

}



// =============================================================================
// ===========================================================================
// Letter (single character)
class Letter implements Shape, Serializable{
  short x1, y1, x2, y2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  Letter(){};

  // =============================================================================
  Letter(short X, short Y, int c, short t){
    x1=X;
    y1=Y;
    RGB=c;
    thick=t;
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    int[][]letter=Alphabet.getLetter((char)thick);
    for (int y=letter.length-1; y>-1;y--){
      for (int x=0; x<letter[0].length;x++){
        if (letter[y][x]!=0)
          Line.pixel(matrix,width,x1+x,y1+y,RGB);
      }
    }
  }

  // =============================================================================
  public void setx2y2(short xx2, short yy2){
    x2=xx2;y2=yy2;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    x1=xx1;
    y1=yy1;
    RGB=cc;
    thick=tt;
    x2=xx2;y2=yy2;
  };
}


// =============================================================================
// ===========================================================================
// LetterBox (texttool rectangle shape)
class LetterBox implements Shape, Serializable{
  short x1, y1, x2, y2;
  int RGB;
  short thick;
  public int ID=0;

  // Set unique ID
  public void setID(int i){ID=i;}

  // Get unique ID
  public int getID(){return ID;}

  // =============================================================================
  LetterBox(){
      x2=y2=-1;
  };

  // =============================================================================
  LetterBox(short X, short Y, int c, short t){
    x1=X;
    y1=Y;
    RGB=c;
    thick=t;
    x2=y2=-1;
  }

  // =============================================================================
  public void draw(int[] matrix, short width){
    int ox=0;
    int oy=3;
    Line.drawthickline(matrix,width,RGB,(short)(ox+x2-3),(short)(oy+y2-10),(short)(ox+x2+3),(short)(oy+y2-10), (short)1);
    Line.drawthickline(matrix,width,RGB,(short)(ox+x2-3),(short)(oy+y2+3),(short)(ox+x2+3),(short)(oy+y2+3), (short)1);
    Line.drawthickline(matrix,width,RGB,(short)(ox+x2+3),(short)(oy+y2-10),(short)(ox+x2+3),(short)(oy+y2+3), (short)1);
    Line.drawthickline(matrix,width,RGB,(short)(ox+x2-3),(short)(oy+y2+3),(short)(ox+x2-3),(short)(oy+y2-10), (short)1);
  }

  // =============================================================================
  public void setx2y2(short xx2, short yy2){
    x2=xx2;y2=yy2;
  };

  // =============================================================================
  public void init(short xx1,short yy1,int cc, short tt, short xx2, short yy2){
    x1=xx1;
    y1=yy1;
    RGB=cc;
    thick=tt;
    x2=xx1;y2=yy1;
  };
}

