/*
Drawboard - Java applet used to make graphical teleconferences
Copyright (C) 2001  Tomek "TomasH" Zielinski, tomash@fidonet.org.pl

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package drawboard;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;

public class Main extends Applet implements Runnable{
  final static short RIGHT_MARGIN=100;    // Menu width
  final static short BOTTOM_MARGIN=50;    // Colorpicker height
  private short DIM_X, DIM_Y;             // Window dimensions
  private int PORT=7904;                  // Port number
  int ID=0;

  private volatile MemoryImageSource bgSource;  // Offline image generator
  volatile Image bgImage;                       // Offline image
  volatile int matrix[];                        // Offline image array
  volatile int matrix2[];                       // Temporary offline image array
  Thread repainter;                             // Image repaint thread
  boolean listenersInitialized=false;           // Initializes event listeners when images loaded

  Board board;            // Blackboard instance
  ColorPicker picker;     // Colorpicker instance
  Menu menu;              // Menu instance
  Communicator comm=null; // Communication module

  int initBgColor=0xFFA6A8F8;             // initial background color (before server connection)
  int initPenColor=0xFF10139E;            // initial pen color
  int counterColor=Color.gray.getRGB();   // menu people counter color
  int menuBgColor=Color.black.getRGB();   // menu background color
  int menuEmptyThColor=Color.darkGray.getRGB(); // menu empty clipboard color

  Alphabet a=new Alphabet();  // fake initialization, IE requires one instance of
                              // any class to use its static methods

// =============================================================================
// =============================================================================
  public void init(){

    // Generate unique ID
    ID=(int)(Math.random()*1000000000)*1000000+(
      (int)(System.currentTimeMillis()&0xfffff));

    // Determine applet size
    try{
      DIM_X = (short)(Integer.parseInt( getParameter( "width" )));
      DIM_Y = (short)(Integer.parseInt( getParameter( "height")));
      System.out.println("Applet window dimensions: width = " + DIM_X + ", height = " + DIM_Y);
    }catch(Exception e){DIM_X=600; DIM_Y=450;} // usefult only when developing with JBuilder
    this.setSize(DIM_X,DIM_Y);

    //Read runtime parameters
    readParameters();

    // Initialize arrays
    matrix=new int[(DIM_X-RIGHT_MARGIN)*(DIM_Y-BOTTOM_MARGIN)];
    matrix2=new int[(DIM_X-RIGHT_MARGIN)*(DIM_Y-BOTTOM_MARGIN)];
    for (int i=0; i<matrix.length; i++){
      matrix[i]=initBgColor;
      matrix2[i]=initBgColor;
    }

    // Initialize image generator and start repainter
    bgSource=new MemoryImageSource((DIM_X-RIGHT_MARGIN),(DIM_Y-BOTTOM_MARGIN), ColorModel.getRGBdefault(), matrix2, 0, (DIM_X-RIGHT_MARGIN));
    bgImage = createImage(bgSource);
    repainter=new Thread(this);
    repainter.setPriority(Thread.MIN_PRIORITY);
    repainter.start();

    this.setLayout(null);

    // Menu instance
    menu=new Menu(getCodeBase(), this,getParameter("skindef"));
    menu.setSize(RIGHT_MARGIN,DIM_Y);
    menu.setCursor(new Cursor(Cursor.HAND_CURSOR));
    menu.setBounds(DIM_X-RIGHT_MARGIN,0,RIGHT_MARGIN,DIM_Y);
    add(menu);
    menu.setVisible(false);

    /// picker
    picker=new ColorPicker(DIM_X-RIGHT_MARGIN,BOTTOM_MARGIN);
    picker.setSize(DIM_X-RIGHT_MARGIN,BOTTOM_MARGIN);
    picker.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
    picker.setBounds(0,DIM_Y-50,DIM_X-100,50);
    add(picker);
    picker.setVisible(false);

    // board
    board=new Board((short)(DIM_X-RIGHT_MARGIN),(short)(DIM_Y-BOTTOM_MARGIN), this);

    // Communicator
    comm=new Communicator(board, getCodeBase(), menu, PORT, this);

    // Make links between objects
    board.setLinks(picker, menu, comm);
    menu.setBoard(board);
    menu.setApplet(this);
    menu.setAppletContext(this.getAppletContext());
    menu.setCommunicator(comm);
    comm.start();

    // Set extra data
    picker.setNewColor(initPenColor);  // initial pen color
    menu.setBgColor(menuBgColor);
    menu.setCounterColor(counterColor);
    menu.setEmptyThColor(menuEmptyThColor);

    setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));

    try{
      menu.tracker.waitForAll(100);
    }catch(InterruptedException e){}
  }


  // =============================================================================
  // Close all connections and exit
  public void destroy(){
    comm.communicatorclosing=true;
  }

  // =============================================================================
  // Subthread, repaints image
  public void run(){
    while(true){
      repaint();
      try{
        Thread.sleep(50);
      }
      catch (InterruptedException e){
        System.out.println("repaint thread interrupted");
      }
    }
  }

  // =============================================================================
  // Blackboard area repaint
  public void paint(Graphics g){
   if (!listenersInitialized){
      g.drawString("Drawboard",10,25);
      g.drawString("(C) by Tomasz 'TomasH' Zielinski",10,40);
      g.drawString("E-mail: tomash@fidonet.org.pl",10,55);
      g.drawString("WWW: http://drawboard.sf.net",10,70);
      g.drawString("Loading graphics...",10,90);
    if (menu.tracker.isErrorAny())
      g.drawString("LOADING GRAPHICS ERROR",10,130);
    }else{
      g.setColor(new Color(menuBgColor));
      g.fillRect(0,0,DIM_X,DIM_Y);
    }
  }

  // =============================================================================
  // Graphics handle
  public void update(Graphics g){
    if (!menu.tracker.checkAll(true) || menu.tracker.isErrorAny()){

        paint(g);
    }else{
      if (!listenersInitialized) setListeners(g);
      board.paint(g);
    }
  }

  // =============================================================================
  // After menu images loaded
  public void setListeners(Graphics g){
    listenersInitialized=true;
    picker.setVisible(true);
    picker.addMouseListener(picker);
    picker.addMouseMotionListener(picker);
    menu.setVisible(true);
    menu.addMouseListener(menu);
    menu.addMouseMotionListener(menu);
    addMouseListener(board);
    addMouseMotionListener(board);

    enableEvents(AWTEvent.KEY_EVENT_MASK);
    addKeyListener(board);
    addFocusListener(board);
    requestFocus();
  }

  // =============================================================================
  // Read HTML-embedded parameters
  void readParameters(){
    try{
      if (getParameter("port")!=null){
        PORT=Integer.parseInt(getParameter("port"));
        System.out.println("Non-default port number defined: "+PORT);
      }
    }catch(NumberFormatException e){System.out.println("Wrong port number, reverting to default value: "+PORT);}

    try{
      if (getParameter("bgcolor")!=null)
        initBgColor=(255<<24)|
          ((Integer.parseInt(getParameter("bgcolor").substring(0,2),16))<<16)|
          ((Integer.parseInt(getParameter("bgcolor").substring(2,4),16))<<8)|
          ((Integer.parseInt(getParameter("bgcolor").substring(4,6),16)));
    }catch(NumberFormatException e){System.out.println("Wrong bgcolor definition, reverting to default value:"+
        " R"+((initBgColor>>16)&0xFF)+" G"+((initBgColor>>8)&0xFF)+" B"+((initBgColor)&0xFF));}

    try{
      if (getParameter("pencolor")!=null)
        initPenColor=(255<<24)|
          (Integer.parseInt(getParameter("pencolor").substring(0,2),16))<<16|
          (Integer.parseInt(getParameter("pencolor").substring(2,4),16))<<8|
          (Integer.parseInt(getParameter("pencolor").substring(4,6),16));
    }catch(NumberFormatException e){System.out.println("Wrong pen color definition, reverting to default value:"+
        " R"+((initPenColor>>16)&0xFF)+" G"+((initPenColor>>8)&0xFF)+" B"+((initPenColor)&0xFF));}

    try{
      if (getParameter("countercolor")!=null)
        counterColor=(255<<24)|
          (Integer.parseInt(getParameter("countercolor").substring(0,2),16))<<16|
          (Integer.parseInt(getParameter("countercolor").substring(2,4),16))<<8|
          (Integer.parseInt(getParameter("countercolor").substring(4,6),16));
    }catch(NumberFormatException e){System.out.println("Wrong countercolor definition, reverting to default value:"+
        " R"+((counterColor>>16)&0xFF)+" G"+((counterColor>>8)&0xFF)+" B"+((counterColor)&0xFF));}

    try{
      if (getParameter("menubgcolor")!=null)
        menuBgColor=(255<<24)|
          (Integer.parseInt(getParameter("menubgcolor").substring(0,2),16))<<16|
          (Integer.parseInt(getParameter("menubgcolor").substring(2,4),16))<<8|
          (Integer.parseInt(getParameter("menubgcolor").substring(4,6),16));
    }catch(NumberFormatException e){System.out.println("Wrong menubgcolor definition, reverting to default value:"+
        " R"+((menuBgColor>>16)&0xFF)+" G"+((menuBgColor>>8)&0xFF)+" B"+((menuBgColor)&0xFF));}

    try{
      if (getParameter("emptythumbnailcolor")!=null)
        menuEmptyThColor=(255<<24)|
          (Integer.parseInt(getParameter("emptythumbnailcolor").substring(0,2),16))<<16|
          (Integer.parseInt(getParameter("emptythumbnailcolor").substring(2,4),16))<<8|
          (Integer.parseInt(getParameter("emptythumbnailcolor").substring(4,6),16));
    }catch(NumberFormatException e){System.out.println("Wrong emptythumbnailcolor definition, reverting to default value:"+
        " R"+((menuBgColor>>16)&0xFF)+" G"+((menuBgColor>>8)&0xFF)+" B"+((menuBgColor)&0xFF));}

  }

}